/*
    Copyright 2008 David Clausen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>

#define MAX_DUTY 110
#define VOLTAGE_SENSITIVITY 100

#define START_VOLTAGE_GREEN 680
#define MAX_VOLTAGE_GREEN 800

#define START_VOLTAGE_RED 480
#define MAX_VOLTAGE_RED 600

#define START_VOLTAGE_BLUE 680
#define MAX_VOLTAGE_BLUE 800

#define MAX_INCREASE 5

static volatile int targetVoltageGreen;
static volatile int targetVoltageRed;
static volatile int targetVoltageBlue;
static volatile char powered = 0;

static void adjust_power_green(int voltage) {
  static volatile int duty = 0;
  int newDuty;
  if (targetVoltageGreen > (voltage + MAX_INCREASE)) {
    duty += MAX_INCREASE;
  } else {
    duty += targetVoltageGreen - voltage;
  }
  if (! powered) {
    duty = 0;
  }
  newDuty = duty / VOLTAGE_SENSITIVITY;
  if (newDuty < 0) {
    duty = newDuty = 0;
  } else if (newDuty > MAX_DUTY) {
    newDuty = MAX_DUTY;
    duty = MAX_DUTY * VOLTAGE_SENSITIVITY;
  }
  OCR0A = newDuty;
}

static void adjust_power_red(int voltage) {
  static volatile int duty = 0;
  int newDuty;
  if (! powered) {
    duty = 0;
  }
  if (targetVoltageRed > (voltage + MAX_INCREASE)) {
    duty += MAX_INCREASE;
  } else {
    duty += targetVoltageRed - voltage;
  }
  newDuty = duty / VOLTAGE_SENSITIVITY;
  if (newDuty < 0) {
    duty = newDuty = 0;
  } else if (newDuty > MAX_DUTY) {
    newDuty = MAX_DUTY;
    duty = MAX_DUTY * VOLTAGE_SENSITIVITY;
  }
  OCR0B = newDuty;
}

static void adjust_power_blue(int voltage) {
  static volatile int duty = 0;
  int newDuty;
  if (! powered) {
    duty = 0;
  }
  if (targetVoltageBlue > (voltage + MAX_INCREASE)) {
    duty += MAX_INCREASE;
  } else {
    duty += targetVoltageBlue - voltage;
  }
  newDuty = duty / VOLTAGE_SENSITIVITY;
  if (newDuty < 0) {
    duty = newDuty = 0;
  } else if (newDuty > MAX_DUTY) {
    newDuty = MAX_DUTY;
    duty = MAX_DUTY * VOLTAGE_SENSITIVITY;
  }
  OCR1A = newDuty;
}

#define STEP 10

static void red_up() {
  if (targetVoltageRed < (MAX_VOLTAGE_RED - STEP)) {
    targetVoltageRed += STEP;
  }
}

static void green_up() {
  if (targetVoltageGreen < (MAX_VOLTAGE_GREEN - STEP)) {
    targetVoltageGreen += STEP;
  }

}

static void blue_up() {
  if (targetVoltageBlue < (MAX_VOLTAGE_BLUE - STEP)) {
    targetVoltageBlue += STEP;
  }
}

static void red_down() {
  if (targetVoltageRed >= STEP) {
    targetVoltageRed -= STEP;
  }
}

static void green_down() {
  if (targetVoltageGreen >= STEP) {
    targetVoltageGreen -= STEP;;
  }
}

static void blue_down() {
  if (targetVoltageBlue >= STEP) {
    targetVoltageBlue -= STEP;
  }
}

static void power_toggle() {
  powered ^= 1;
  if (powered) {
    // connect ground pin
    PORTD |= _BV(PD2);

    // start PWM on OC0A and OC0B, fast pwm, top = 0xff
    TCCR0B = 0;
    TCNT0 = 0; // reset timer0 to 0
    OCR0A = 0;
    OCR0B = 0;
    TCCR0A = _BV(COM0A1) | _BV(COM0B1) | _BV(WGM01) | _BV(WGM00);
    TCCR0B = _BV(CS00);
    
    // start PWM on OC1A, 8 bit, fast pwm, top = 0xff
    TCCR1B = 0;
    TCNT1 = 0;
    OCR1A = 0;
    OCR1B = 0;
    TCCR1A = _BV(COM1A1) | _BV(WGM10);
    TCCR1B = _BV(WGM12) | _BV(CS10);
    
  } else {
    // stop pwm timers
    TCCR0B = 0;
    TCCR0A = 0;
    TCCR1B = 0;
    TCCR1A = 0;

    // make sure the pwm outputs are low!
    PORTD &= ~(_BV(PD5) | _BV(PD6));
    PORTB &= ~_BV(PB1);

    // disconnect ground pin
    PORTD &= ~_BV(PD2);
  }
}

#define DEBOUNCE 100

static void read_button_state() {
  static volatile int cRed = 0;
  static volatile int cGreen = 0;
  static volatile int cBlue = 0;
  static volatile int cDown = 0;
  static volatile int cUp = 0;
  int red = !(PIND & _BV(PD3));
  int green = !(PIND & _BV(PD4));
  int blue = !(PIND & _BV(PD7));
  int down = !(PINB & _BV(PB0));
  int up = !(PINB & _BV(PB2));

  if (red) {
    cRed++;
    red = (cRed >= DEBOUNCE);
    if (red) {
      cRed = DEBOUNCE;
    }
  } else {
    cRed = 0;
  }

  if (green) {
    cGreen++;
    green = (cGreen >= DEBOUNCE);
    if (green) {
      cGreen = DEBOUNCE;
    }
  } else {
    cGreen = 0;
  }

  if (blue) {
    cBlue++;
    blue = (cBlue >= DEBOUNCE);
    if (blue) {
      cBlue = DEBOUNCE;
    }
  } else {
    cBlue = 0;
  }

  if (down) {
    cDown++;
    down = cDown == DEBOUNCE;
    if (cDown > DEBOUNCE) {
      cDown = DEBOUNCE;
    }
  } else {
    cDown = 0;
  }

  if (up) {
    cUp++;
    up = cUp == DEBOUNCE;
    if (cUp > DEBOUNCE) {
      cUp = DEBOUNCE;
    }
  } else {
    cUp = 0;
  }

  if (up && down) {
    if (red) {
      targetVoltageRed = START_VOLTAGE_RED;
    }
    if (blue) {
      targetVoltageBlue = START_VOLTAGE_BLUE;
    }
    if (green) {
      targetVoltageGreen = START_VOLTAGE_GREEN;
    }
  } else if (up) {
    if (! (red || green || blue)) {
      power_toggle();
    }
    if (red) {
      red_up();
    }
    if (green) {
      green_up();
    }
    if (blue) {
      blue_up();
    }
  } else if (down) {
    if (red) {
      red_down();
    }
    if (green) {
      green_down();
    }
    if (blue) {
      blue_down();
    }
  }


}


ISR(ADC_vect) {
  char which = ADMUX & 0x7;
  int adcw = ADCW;
  ADMUX &= ~ 0x7;
  if (which == 5) {
    ADMUX |= 4;
  } else if (which == 4) {
    ADMUX |= 3;
  } else {
    ADMUX |= 5; // green
  }
  ADCSRA |= _BV(ADSC); // start next conversion
  if (which == 5) {
    adjust_power_green(adcw);
  } else if (which == 4) {
    adjust_power_red(adcw);
  } else {
    adjust_power_blue(adcw);
  }
  read_button_state();
}

static volatile int k = 0;

static int my_delay() {
  int i, j;
  k = 0;
  for (i = 0; i < 1000; i++) {
    for (j = 0; j < 1000; j++) {
      k++;
    }
  }
  return k;
}

int main(void) {

  PORTD = 0;
  PORTB = 0;

  // setup power control mosfet
  DDRD |= _BV(PD2);
  
  // setup button inputs and enable pull-up resistors
  DDRD &= ~(_BV(PD3) | _BV(PD4) | _BV(PD7));
  PORTD |= _BV(PD3) | _BV(PD4) | _BV(PD7);
  DDRB &= ~(_BV(PB0) | _BV(PB2));
  PORTB |= _BV(PB0) | _BV(PB2);

  // set pwm outputs
  DDRD |= _BV(PD6) | _BV(PD5);
  DDRB |= _BV(PB1);
    
  // start ADC
  DDRC = 0;
  PORTC = 0;
  ADMUX = _BV(REFS0) + 5;
  ADCSRB = 0;
  ADCSRA = _BV(ADEN) | _BV(ADSC) | _BV(ADIE) | _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0);

  sei();
  
  my_delay();

  targetVoltageRed = START_VOLTAGE_RED;
  targetVoltageGreen = 0;
  targetVoltageBlue = 0;

  power_toggle();

  my_delay();

  targetVoltageRed = 0;
  targetVoltageGreen = START_VOLTAGE_GREEN;
  targetVoltageBlue = 0;

  my_delay();

  targetVoltageRed = 0;
  targetVoltageGreen = 0;
  targetVoltageBlue = START_VOLTAGE_BLUE;

  my_delay();

  power_toggle();

  targetVoltageRed = START_VOLTAGE_RED;
  targetVoltageGreen = START_VOLTAGE_GREEN;
  targetVoltageBlue = START_VOLTAGE_BLUE;

  return 0;
}

